/* This file is part of Mailfromd.
   Copyright (C) 2007, 2008 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "pies.h"

#define TYPE_COMPONENT 0
#define TYPE_RETR      1

struct prog
{
  struct prog *next;
  int type;
  pid_t pid;             /* PID */
  char *tag;             /* Entry tag (for diagnostics purposes) */
  char **depend;
  union
  {
    struct
    {
      int argc;
      char **argv;           /* Command line arguments */
      char **env;            /* Environment */
      int retr[2];
      char *rmfile;          /* Rmfile location */
      struct pies_privs_data privs;
      mode_t umask;
      time_t timestamp;      /* Time of last startup */
      size_t count;          /* Number of failed starts since timestamp */
      int disabled;          /* 1 if this entry is disabled */
    } p;
    
    struct
    {
      struct prog *master;
    } r;
  } v;
};

#define IS_PROG(p) ((p)->type == TYPE_COMPONENT)

static struct prog *proghead, *progtail;

static void prog_stop (struct prog *prog, int sig);

void
link_prog (struct prog *pp, int prepend)
{
  if (prepend)
    {
      pp->next = proghead;
      proghead = pp;
      if (!progtail)
	progtail = pp;
    }
  else
    {
      pp->next = NULL;
      if (progtail)
	progtail->next = pp;
      else
	proghead = pp;
      progtail = pp;
    }
}

void
register_retr (int type, struct prog *master, pid_t pid)
{
  struct prog *pp;
  static char *retrstr[2] = { "stdout", "stderr" };
  size_t taglen = strlen (master->tag) + 1 + strlen (retrstr[type]);
  char *tag = xmalloc (taglen + 1);
  
  strcpy (tag, master->tag);
  strcat (tag, "/");
  strcat (tag, retrstr[type]); 
  
  for (pp = proghead; pp; pp = pp->next)
    if (pp->type == TYPE_RETR && pp->v.r.master == master
	&& strcmp (pp->tag, tag) == 0)
      {
	free (tag);
	prog_stop (pp, SIGKILL);
	break;
      }

  if (!pp)
    {
      pp = xmalloc (sizeof (*pp) + taglen + 1);
      memset (pp, 0, sizeof (*pp));
      pp->type = TYPE_RETR;
      pp->tag = (char *) (pp + 1);
      strcpy (pp->tag, tag);
      free (tag);
      pp->v.r.master = master;
      link_prog (pp, 1);
    }
  
  pp->pid = pid;
}

void
register_prog (struct component *comp)
{
  struct prog *p, *newp;
  char *pstr;
  int i;
  size_t size = 0;
  int dep_all = 0;
  int depc = 0;
  
  if (comp->depend)
    {
      if (comp->depend[0] && comp->depend[1] == NULL)
	{
	  if (strcmp (comp->depend[0], "all") == 0)
	    {
	      dep_all = 1;
	      for (p = proghead; p; p = p->next)
		if (p->type == TYPE_COMPONENT)
		  {
		    depc++;
		    size += strlen (p->tag) + 1;
		  }
	    }
	  else if (strcmp (comp->depend[0], "none") == 0)
	    comp->depend = NULL;
	}

      if (depc == 0)
	for (depc = 0; comp->depend[depc]; depc++)
	  ;
      size += sizeof (comp->depend[0]) * (depc + 1);
    }
  
  size += sizeof (*newp)
          + (comp->tag ? (strlen (comp->tag) + 1) : 0);
          + (comp->rmfile ? (strlen (comp->rmfile) + 1) : 0);

  for (i = 0; comp->argv[i]; i++)
    ;
  
  newp = xzalloc (size);
  newp->type = TYPE_COMPONENT;
  newp->pid = 0;
  newp->v.p.argc = i;
  newp->v.p.argv = comp->argv;
  newp->v.p.env = comp->env;
  
  pstr = (char*) (newp + 1);
  
  if (comp->tag)
    {
      newp->tag = strcpy (pstr, comp->tag);
      pstr += strlen (pstr) + 1;
    }
  else
    newp->tag = newp->v.p.argv[0];

  if (comp->rmfile)
    {
      newp->v.p.rmfile = strcpy (pstr, comp->rmfile);
      pstr += strlen (pstr) + 1;
    } 
  else
    newp->v.p.rmfile = NULL;
  
  if (comp->depend)
    {
      newp->depend = (char**) pstr;
      pstr = (char*) (newp->depend + depc + 1);
      if (dep_all)
	{
	  depc = 0;
	  for (p = proghead; p; p = p->next)
	    if (p->type == TYPE_COMPONENT)
	      {
		newp->depend[depc++] = pstr;
		strcpy (pstr, p->tag);
		pstr += strlen (pstr) + 1;
	      }
	}
      else
	{
	  for (depc = 0; comp->depend[depc]; depc++)
	    {
	      newp->depend[depc] = pstr;
	      strcpy (pstr, comp->depend[depc]);
	      pstr += strlen (pstr) + 1;
	    }
	}
      newp->depend[depc] = NULL;
    } 
  else
    newp->depend = NULL;
  
  newp->v.p.retr[0] = comp->retr[0];
  newp->v.p.retr[1] = comp->retr[1];

  newp->v.p.privs = comp->privs;
  newp->v.p.umask = comp->umask;
  
  link_prog (newp, 0);
}

static struct prog *
find_prog (pid_t pid)
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    if (prog->pid == pid)
      break;
  return prog;
}

size_t
progman_running_count ()
{
  size_t size = 0;
  struct prog *prog;
    
  for (prog = proghead; prog; prog = prog->next)
    if (prog->pid > 0)
      size++;
  return size;
}

RETSIGTYPE
retr_exit (int sig)
{
  exit (0);
}

int
open_retranslator (struct prog *master, int type)
{
  int p[2];
  FILE *fp;
  char *buf = NULL;
  size_t size = 0;
  pid_t pid;
  
  if (master->v.p.retr[type] == -1)
    return -1;
  pipe (p);
  switch (pid = fork ())
    {
    case 0:
      /* Retranslator process */

      log_setup (0);
      signal_setup (retr_exit);
      
      close (p[1]);
      fp = fdopen (p[0], "r");
      if (fp == NULL)
	exit (1);
      
      openlog (master->tag, LOG_PID, mu_log_facility);
      
      while (getline (&buf, &size, fp) > 0)
	syslog (master->v.p.retr[type], "%s", buf);
      exit (0);
      
    case -1:
      mu_diag_output (MU_DIAG_CRIT,
		      _("cannot run retranslator `%s': fork failed: %s"),
		      master->tag, mu_strerror (errno));
      return -1;

    default:
      register_retr (type, master, pid);
	
      close (p[0]);
      return p[1];
    }
}

extern char **environ;

static char *
find_env (char *name)
{
  int nlen = strlen (name);
  int i;

  for (i = 0; environ[i]; i++)
    {
      char *p = strchr (environ[i], '=');
      int elen = p - environ[i];
      if (elen == nlen && memcmp (name, environ[i], nlen) == 0)
	return environ[i];
    }
  return NULL;
}

static void
env_setup (char **env)
{
  char **old_env = environ;
  char **new_env;
  int count, i, n;
  
  if (!env)
    return;

  if (strcmp (env[0], "-") == 0)
    {
      old_env = NULL;
      env++;
    }

  /* Count new environment size */
  count = 0;
  if (old_env)
    for (; old_env[count]; count++)
      ;
    
  for (; env[count]; count++)
    ;

  /* Allocate the new environment. */
  new_env = xcalloc (count + 1, sizeof new_env[0]);

  /* Populate the environment. */
  n = 0;
  
  if (old_env)
    for (i = 0; old_env[i]; i++)
      new_env[n++] = old_env[i];

  for (i = 0; env[i]; i++)
    {
      if (strchr (env[i], '='))
	new_env[n++] = env[i];
      else
	{
	  char *p = find_env (env[i]);
	  if (p)
	    new_env[n++] = p;
	}
    }
  new_env[n] = NULL;
  /* FIXME: The old environment is not freed. Should it be? */
  environ = new_env; 
}

static void
prog_start (struct prog *prog)
{
  int i;
  pid_t pid;
  time_t now;
  int retr[2];
  
  time (&now);

  if (prog->v.p.timestamp + TESTTIME > now)
    prog->v.p.count++;
  else
    {
      prog->v.p.count = 0;
      prog->v.p.timestamp = now;
    }

  if (prog->v.p.count > MAXSPAWN)
    {
      int old_alarm;
      
      mu_error (_("%s is respawning too fast, disabled for %d minutes"),
		prog->tag, SLEEPTIME / 60);
      prog->v.p.timestamp = now;
      prog->v.p.disabled = 1;
      
      old_alarm = alarm (0);
      if (old_alarm > SLEEPTIME || old_alarm <= 0)
	old_alarm = SLEEPTIME;
      alarm (old_alarm);
      return;
    }
    
  mu_diag_output (MU_DIAG_DEBUG, _("starting %s"), prog->tag);
  
  if (prog->v.p.rmfile)
    {
      if (unlink (prog->v.p.rmfile) && errno != ENOENT)
	mu_error (_("%s: cannot remove file `%s': %s"),
		  prog->tag, prog->v.p.rmfile, mu_strerror (errno));
    }
      
  retr[RETR_OUT] = open_retranslator (prog, RETR_OUT);
  retr[RETR_ERR] = open_retranslator (prog, RETR_ERR);
  
  switch (pid = fork ())
    {
      /* The child branch.  */
    case 0:
       if (retr[RETR_OUT] == -1)
	 {
	   close (1);
	   open ("/dev/null", O_RDWR);
	 }
       else if (retr[RETR_OUT] != 1)
	 {
	   close (1);
	   dup2 (retr[RETR_OUT], 1);
	 }

       if (retr[RETR_ERR] == -1)
	 {
	   close (2);
	   open ("/dev/null", O_RDWR);
	 }
       else if (retr[RETR_ERR] != 1)
	 {
	   close (2);
	   dup2 (retr[RETR_ERR], 2);
	 }

       /* Close unneded descripitors */
       for (i = getmaxfd (); i > 2; i--)
	 close (i);

       signal_setup (SIG_DFL);
       env_setup (prog->v.p.env);
       priv_setup (&prog->v.p.privs);
       if (prog->v.p.umask)
	 umask (prog->v.p.umask);

       execvp (prog->v.p.argv[0], prog->v.p.argv);
       openlog (MU_LOG_TAG(), LOG_PID, mu_log_facility);
       syslog (LOG_CRIT, _("cannot start `%s': %s"), prog->tag,
	       mu_strerror (errno));
       exit (1);

    case -1:
      mu_diag_output (MU_DIAG_CRIT,
		      _("cannot run `%s': fork failed: %s"),
		      prog->tag, mu_strerror (errno));
      break;

    default:
      prog->pid = pid;
    }
}

void
progman_start ()
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    if (IS_PROG (prog))
	prog_start (prog);
}

void
progman_wake_disabled ()
{
  struct prog *prog;

  for (prog = proghead; prog; prog = prog->next)
    if (IS_PROG (prog) && prog->v.p.disabled)
      {
	prog->v.p.disabled = 0;
	prog->v.p.count = 0;
	prog->v.p.timestamp = 0;
	prog_start (prog);
      }
}

static void
prog_stop (struct prog *prog, int sig)
{
  if (prog->pid > 0)
    {
      mu_diag_output (MU_DIAG_DEBUG, "Stopping %s (%lu)",
		      prog->tag, (unsigned long) prog->pid);
      kill (prog->pid, sig);
    }
}  

static void
prog_stop_recursive (struct prog *prog, int sig)
{
  if (!prog)
    return;
  prog_stop_recursive (prog->next, sig);
  prog_stop (prog, sig);
}

void
progman_stop ()
{
  unsigned long i;
  
  prog_stop_recursive (proghead, SIGTERM);
  for (i = 0; i < shutdown_timeout; i++)
    {
      progman_cleanup (1);
      if (progman_running_count () == 0)
	return;
      sleep (1);
    }
  prog_stop_recursive (proghead, SIGKILL);
}

static void
print_status (char *tag, pid_t pid, int status, int expect_term)
{
  if (WIFEXITED (status))
    {
      if (WEXITSTATUS (status) == 0)
	mu_diag_output (MU_DIAG_DEBUG,
			_("%s (%lu) exited successfully"),
			tag, (unsigned long) pid);
      else
	mu_diag_output (MU_DIAG_ERR,
			_("%s (%lu) failed with status %d"),
			  tag, (unsigned long) pid,
			  WEXITSTATUS (status));
    }
  else if (WIFSIGNALED (status))
    {
      int prio;

      if (expect_term && WTERMSIG (status) == SIGTERM)
	prio = MU_DIAG_DEBUG;
      else
	prio = MU_DIAG_ERR;

      mu_diag_output (prio,
		      _("%s (%lu) terminated on signal %d"),
		      tag, (unsigned long) pid,
		      WTERMSIG (status));
    }
  else if (WIFSTOPPED (status))
    mu_diag_output (MU_DIAG_ERR,
		    _("%s (%lu) stopped on signal %d"),
		    tag, (unsigned long) pid,
		    WSTOPSIG (status));
#ifdef WCOREDUMP
  else if (WCOREDUMP (status))
    mu_diag_output (MU_DIAG_ERR,
		    _("%s (%lu) dumped core"),
		    tag, (unsigned long) pid);
#endif
  else
    mu_diag_output (MU_DIAG_ERR,
		    _("%s (%lu) terminated with unrecognized status"),
		    tag, (unsigned long) pid);
}

void
prog_stop_dependent (struct prog *prog)
{
  struct prog *p;
  
  for (p = proghead; p; p = p->next)
    if (p->depend)
      {
	int i;
	for (i = 0; p->depend[i]; i++)
	  if (strcmp (prog->tag, p->depend[i]) == 0)
	    prog_stop (p, SIGTERM);
      }
}

void
progman_cleanup (int expect_term)
{
  pid_t pid;
  int status;
  while ((pid = waitpid (-1, &status, WNOHANG)) > 0)
    {
      struct prog *prog = find_prog (pid);
      if (!prog)
	{
	  mu_diag_output (MU_DIAG_NOTICE,
			  _("unknown child %lu finished"),
			  (unsigned long) pid);
	  continue;
	}
      print_status (prog->tag, prog->pid, status, expect_term);
      prog->pid = 0;
      if (IS_PROG (prog))
	{
	  prog_stop_dependent (prog);
	  if (!expect_term)
	    prog_start (prog);
	}
    }
}

void
progman_stop_component (const char *name)
{
  struct prog *prog;

  mu_diag_output (MU_DIAG_INFO, _("stopping component `%s'"), name);
  for (prog = proghead; prog; prog = prog->next)
    if (IS_PROG (prog) && strcmp (prog->tag, name) == 0)
      {
	if (prog->v.p.disabled)
	  mu_diag_output (MU_DIAG_INFO,
			  _("stopping component `%s': component not started"),
		  name);
	else
	  prog_stop (prog, SIGTERM);
      }
}

void
progman_dump_stats (const char *filename)
{
  FILE *fp;
  struct prog *prog;
  char *s;
  int rc;
  
  mu_diag_output (MU_DIAG_INFO, _("dumping statistics to `%s'"), filename);
  fp = fopen (filename, "w");
  if (!fp)
    {
      mu_error (_("cannot open file `%s' for writing: %s"),
		filename, mu_strerror (errno));
      return;
    }

  for (prog = proghead; prog; prog = prog->next)
    {
      switch (prog->type)
	{
	case TYPE_COMPONENT:
	  fprintf (fp, "component %s ",
		   prog->tag);
	  if (prog->pid)
	    fprintf (fp, "%lu", (unsigned long) prog->pid);
	  else if (prog->v.p.disabled)
	    {
	      char buf[48];
	      time_t t = prog->v.p.timestamp + SLEEPTIME;
	      strftime (buf, sizeof buf, "%c",
			localtime (&t));
	      fprintf (fp, _("[disabled; scheduled for %s]"), buf);
	    }
	  else
	    fprintf (fp, "[not running]");
	  if (rc = mu_argcv_string (prog->v.p.argc, prog->v.p.argv, &s))
	    {
	      mu_error (_("cannot convert argument list: %s"),
			mu_strerror (rc));
	    }
	  else
	    {
	      fprintf (fp, " %s", s);
	      free (s);
	    }
	  fputc ('\n', fp);
	  break;
	  
	case TYPE_RETR:
	  fprintf (fp, _("retranslator %s %lu\n"), prog->tag,
		   (unsigned long) prog->pid);
	}
    }
  fclose (fp);
}

/*
 Local Variables:
 c-file-style: "gnu"
 End:
*/
/* EOF */
